/* GtkEditor - a source editor widget for GTK
 * Copyright (C) 1998, 2000 Thomas Mailund.
 *
 * The editor widget was written by Thomas Mailund, so bugs should be
 * reported to <mailund@daimi.au.dk>, not the gtk ppl.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

#include "gtkeditor.h"
#include "internal.h"
#include "edit.h"
#include "scanner.h"

#define NOT_DONE 1
#define DONE 1 << 1

/* pointer to current editor being analized */
static GtkEditor *g_editor;

static const char *g_insertion_text;

/* tells if we are done yet */
static int g_lex_flag;

/* points to character position in the new text */
static int g_char_pos;

/* points to current position in the new text */
static unsigned int g_pos;

/* length of the buffer plus inserted text */
static unsigned int g_length;

static GList *g_prev_token;

/*
 * prototypes
 */
static int            next_char            ();
static void           new_token            (guint16         id,
					    char           *text,
					    guint16         length);
					
/* gtk_editor_compile_scanner

   A flex scanner is generated to search for the highlighted
   entries. The scanner is compiled into a shared library which can be
   loaded at a later time.
*/
GtkEditorScanner *
gtk_editor_compile_scanner (const char *name,
			    GList *entries)
{
  GtkEditorScanner *scanner;
  GtkEditorTokenEntry *entry;
  FILE *flex;
  GList *cur;
  void *scanner_lib;
  char buffer[128];
  int i = 0;

  sprintf (buffer, "%s.flex", name);
  flex = fopen(buffer, "w+");

  if (!flex)
    return NULL;

  scanner = g_new (GtkEditorScanner, 1);

  /* definitions */
  fprintf (flex, "%%option noyywrap\n");
  fprintf (flex, "%%{\n");
  fprintf (flex, "#include <glib.h>\n");
  fprintf (flex, "static void (*new_token)(guint16 id, char *text, guint16 length);\n");
  fprintf (flex, "static int (*next_char)();\n");
  fprintf (flex, "void token_func(void (*func)(guint16 id, char *text, guint16 length));\n");
  fprintf (flex, "void char_func(int (*func)());\n");
  /* YY_INPUT macro for incremental analysis */
  fprintf (flex, "#define YY_INPUT(buf,result,max_size) \\\n");
  fprintf (flex, "{ \\\n");
  fprintf (flex, "  int c = next_char(); \\\n");
  fprintf (flex, "  result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \\\n");
  fprintf (flex, "}\n");
  fprintf (flex, "%%}\n");

  /* rules */
  fprintf (flex, "%%%%\n");
  for (cur = entries; cur != NULL; cur = cur->next) {
    entry = cur->data;
    fprintf (flex, "%s new_token(%d, yytext, yyleng);\n",
	     entry->pattern, i);
    i++;
  }

  /* user code */
  fprintf (flex, "%%%%\n");
  fprintf (flex, "void token_func(void (*func)(guint16 id, char *text, guint16 length))\n");
  fprintf (flex, "{\n\tnew_token = func;\n}\n");
  fprintf (flex, "void char_func(int (*func)())\n");
  fprintf (flex, "{\n\tnext_char = func;\n}\n");
  fclose(flex);

  /* generate the scanner */
  sprintf (buffer, "flex -o%s.c %s.flex", name, name);
  system(buffer);
  /* compile the scanner */
  sprintf (buffer, "gcc `gtk-config --cflags` -shared -o %s %s.c", name, name);
  system(buffer);

#if 0
  /* copy it */
  sprintf (buffer, "cp %s /usr/lib", name);
  system(buffer);
#endif

  scanner_lib = dlopen(name, RTLD_LAZY);
  if (!scanner_lib) {
    g_warning ("could not open dynamic library.");
    return NULL;
  }

  scanner->token_func = dlsym(scanner_lib, "token_func");
  scanner->char_func = dlsym(scanner_lib, "char_func");
  scanner->lex = dlsym(scanner_lib, "yylex");
  scanner->token_func(new_token);
  scanner->char_func(next_char);

  return scanner;
}

/* gtk_editor_load_scanner

   Loads a previously compiled flex scanner.
*/
GtkEditorScanner *
gtk_editor_load_scanner (const char *file_name)
{
  GtkEditorScanner *scanner;
  void *scanner_lib;

  scanner = g_new(GtkEditorScanner, 1);
  scanner_lib = dlopen(file_name, RTLD_LAZY);
  scanner->token_func = dlsym(scanner_lib, "token_func");
  scanner->char_func = dlsym(scanner_lib, "char_func");
  scanner->lex = dlsym(scanner_lib, "yylex");
  scanner->token_func(new_token);
  scanner->char_func(next_char);

  return scanner;
}

void
gtk_editor_install_scanner (GtkEditor *editor,
			    GtkEditorScanner *scanner,
			    GList *entries)
{
  GtkEditorTokenEntry *entry;
  GList *cur;
  int i = 0;

  editor->scanner = scanner;
  editor->n_token_entries = g_list_length (entries);
  editor->token_entries = g_new (GtkEditorTokenEntry, editor->n_token_entries);

  for (cur = entries; cur != NULL; cur = cur->next) {
    entry = cur->data;
    editor->token_entries[i].name = g_strdup (entry->name);
    editor->token_entries[i].pattern = g_strdup (entry->pattern);
    editor->token_entries[i].precedence = entry->precedence;
    editor->token_entries[i].prop_id = entry->prop_id;
    editor->token_entries[i].begin = -1;
    editor->token_entries[i].n_ends = 0;
    editor->token_entries[i].ends = NULL;
    i++;
  }
}

void
gtk_editor_install_bentries (GtkEditor *editor,	GList *bentries)
{
  GtkEditorBlockEntry *entry;
  GList *cur;
  int *temp;
  int i = 0, j;

  editor->n_block_entries = g_list_length (bentries);
  editor->block_entries = g_new (GtkEditorBlockEntry, editor->n_block_entries);

  for (cur = bentries; cur != NULL; cur = cur->next) {
    entry = cur->data;
    editor->block_entries[i].name = g_strdup (entry->name);
    editor->block_entries[i].begin = entry->begin;
    editor->token_entries[entry->begin].begin = i;
    editor->block_entries[i].n_ends = entry->n_ends;
    editor->block_entries[i].ends = g_new (int, entry->n_ends);
    memcpy(editor->block_entries[i].ends, entry->ends, sizeof(int)*entry->n_ends);
    for (j = 0; j < entry->n_ends; j++) {
      if (editor->token_entries[entry->ends[j]].ends) {
        temp = editor->token_entries[entry->ends[j]].ends;
        editor->token_entries[entry->ends[j]].ends =
          g_new(int, editor->token_entries[entry->ends[j]].n_ends + 1);
        memcpy(editor->token_entries[entry->ends[j]].ends, temp,
               sizeof(int)*editor->token_entries[entry->ends[j]].n_ends);
        editor->token_entries[entry->ends[j]].ends[editor->token_entries[entry->ends[j]].n_ends] = i;
        editor->token_entries[entry->ends[j]].n_ends++;
        g_free(temp);
      } else {
        editor->token_entries[entry->ends[j]].n_ends = 1;
        editor->token_entries[entry->ends[j]].ends = g_new(int, 1);
        editor->token_entries[entry->ends[j]].ends[0] = i;
      }
    }
    editor->block_entries[i].precedence = entry->precedence;
    editor->block_entries[i].nest = entry->nest;
    editor->block_entries[i].prop_id = entry->prop_id;
    i++;
  }
}

GList *
gtk_editor_pentry_new (const gchar *name, const gchar *pattern,
		       int precedence, int prop_id, GList *next)
{
  GList *tmp = g_list_alloc ();
  GtkEditorTokenEntry *new = g_new (GtkEditorTokenEntry, 1);

  new->name = g_strdup (name);
  new->pattern = g_strdup (pattern);
  new->begin = -1;
  new->precedence = precedence;
  new->prop_id = prop_id;

  tmp->data = (gpointer)new;
  tmp->next = next;

  return tmp;
}

GList *gtk_editor_bentry_new (const gchar *name, int begin, int n_ends,
			      int *ends, int precedence, gboolean nest,
			      int prop_id, GList *next)
{
  GList *tmp = g_list_alloc ();
  GtkEditorBlockEntry* new = g_new (GtkEditorBlockEntry, 1);

  new->name = g_strdup(name);
  new->begin = begin;
  new->n_ends = n_ends;
  new->ends = g_new (int, n_ends);
  memcpy(new->ends, ends, sizeof(int)*n_ends);
  new->precedence = precedence;
  new->nest = nest;
  new->prop_id = prop_id;

  tmp->data = (gpointer)new;
  tmp->next = next;

  return tmp;
}

void
_gtk_editor_scan (GtkEditor *editor, int pos,
      const char *insertion_text,
      int length, int insertion_point)
{
  g_prev_token = NULL;
  g_editor = editor;
  g_lex_flag = NOT_DONE;
  g_insertion_point = insertion_point;
  g_length = gtk_text_get_length (GTK_TEXT(editor));
  if(length >= 0) {
    g_insertion_text = insertion_text;
    g_insertion_length = length;
    g_deletion_length = 0;
  } else {
    g_insertion_length = 0;
    g_deletion_length = -length;
  }
  g_length += length;
  g_pos = g_char_pos = pos;
  editor->scanner->lex();
}

static int
next_char()
{
  int c;

  if ((g_lex_flag == DONE) || (g_char_pos >= g_length))
    c = EOF;
  else if (g_insertion_text
           && (g_char_pos >= g_insertion_point)
           && (g_char_pos < g_insertion_point + g_insertion_length))
    c = g_insertion_text[g_char_pos - g_insertion_point];
  else if (g_char_pos >= g_insertion_point)
    c = GTK_TEXT_INDEX(GTK_TEXT(g_editor), g_char_pos - g_insertion_length + g_deletion_length);
  else
    c = GTK_TEXT_INDEX(GTK_TEXT(g_editor), g_char_pos);

  g_char_pos++;
  return c;
}

/* new_token

  This function is called by flex when flex completes a token. It inserts the
  tokens into the global editor with the properties specified. The algorithm
  trys to resuse existing tokens in the editor, and only deletes and reinserts
  text when it has changed properties.

*/
static void
new_token(guint16 id, char *text, guint16 length)
{
  GList *cur_token;
  Token *token;

  if (g_lex_flag == DONE)
    /* nothing to do here */
    return;

  /* check if no more lexing is needed */
  if (g_token) {
    token = g_token->data;
    if (g_pos >= g_insertion_point + g_insertion_length
        && token->id == id
        && token->length == length) {
      g_lex_flag = DONE;
    }
  }

  /* merge token into the token list */
  cur_token = _gtk_editor_merge_token(id, text, length, g_pos);

  /* update the cur token */
  if (g_insertion_point >= g_pos
      && g_insertion_point <= g_pos + length) {
    token = cur_token->data;
    g_editor->cur_token = cur_token;
    g_editor->cur_token_pos = g_pos;
    g_editor->cur_block = g_cur_block;
    g_editor->prev_block = g_prev_block;
    g_editor->next_block = g_next_block;
  }

  g_pos += length;
}

